/*
 * Copyright (C) 2006-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file ipc.c Inter-process communication and shared memory helpers. */

#include "para.h"
#include "error.h"
#include "ipc.h"
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>

/**
 * Define a new mutex.
 *
 * \return The identifier for the new mutex on success, a negative error code
 * on errors.
 *
 * \sa semget(2).
 */
int mutex_new(void)
{
	int ret = semget(IPC_PRIVATE, 1, IPC_CREAT | 0666);
	return ret < 0? -ERRNO_TO_PARA_ERROR(errno) : ret;
}

/**
 * Destroy a mutex.
 *
 * \param id The identifier of the mutex to be destroyed.
 *
 * \return Standard.
 *
 * \sa semctl(2)
 */
int mutex_destroy(int id)
{
	int ret = semctl(id, 0, IPC_RMID);
	return ret < 0? -ERRNO_TO_PARA_ERROR(errno) : 1;
}

static void para_semop(int id, struct sembuf *sops, int num)
{
	do {
		if (semop(id, sops, num) >= 0)
			return;
	} while (errno == EINTR);
	if (errno == EIDRM) {
		PARA_CRIT_LOG("semaphore set %d was removed\n", id);
		return;
	}
	PARA_EMERG_LOG("fatal semop error %s: pid %d\n", strerror(errno),
		(int)getpid());
	exit(EXIT_FAILURE);
}

/**
 * Lock the given mutex.
 *
 * \param id The identifier of the shared memory area to lock.
 *
 * This function either succeeds or aborts.
 *
 * \sa semop(2), struct misc_meta_data.
 */
void mutex_lock(int id)
{
	struct sembuf sops[2] = {
		{
			.sem_num = 0,
			.sem_op = 0,
			.sem_flg = SEM_UNDO
		},
		{
			.sem_num = 0,
			.sem_op = 1,
			.sem_flg = SEM_UNDO
		}
	};
	para_semop(id, sops, 2);
}

/**
 * Unlock a mutex.
 *
 * \param id The identifier of the mutex.
 *
 * This function either succeeds or aborts.
 *
 * \sa semop(2), struct misc_meta_data.
 */
void mutex_unlock(int id)
{
	struct sembuf sops[1] = {
		{
			.sem_num = 0,
			.sem_op = -1,
			.sem_flg = SEM_UNDO
		},
	};
	para_semop(id, sops, 1);
}

/**
 * Create a new shared memory area of given size.
 *
 * \param size The size of the shared memory area to create.
 *
 * \return The id of the shared memory array on success, a negative error
 * code on errors.
 *
 * \sa shmget(2).
 */
int shm_new(size_t size)
{
	int ret = shmget(IPC_PRIVATE, size, IPC_CREAT | IPC_EXCL | 0600);
	return ret < 0 ? -ERRNO_TO_PARA_ERROR(errno) : ret;
}

/**
 * Destroy the given shared memory area.
 *
 * \param id The shared memory identifier.
 *
 * \return The return value of the underlying shmctl() call on success,
 * a negative error code on errors.
 *
 * \sa shmctl(2).
 */
int shm_destroy(int id)
{
	struct shmid_ds shm_desc;
	int ret = shmctl(id, IPC_RMID, &shm_desc);
	return ret < 0? -ERRNO_TO_PARA_ERROR(errno) : ret;
}

/**
 * Attach a shared memory segment.
 *
 * \param id The identifier of the shared memory segment to attach.
 * \param mode Either ATTACH_RO (read only) or ATTACH_RW (read/write).
 * \param result Points to the attached area just attached on success.
 *
 * \return Standard.
 *
 * \sa shmat(2).
 */
int shm_attach(int id, enum shm_attach_mode mode, void **result)
{
	if (mode == ATTACH_RW)
		*result = shmat(id, NULL, 0);
	else
		*result = shmat(id, NULL, SHM_RDONLY);
	return *result == (void *) -1? -ERRNO_TO_PARA_ERROR(errno) : 1;
}

/**
 * Detach a shared memory segment.
 *
 * \param addr The address of the attached segment.
 *
 * \return Standard.
 *
 * \sa shmdt(2).
 */
int shm_detach(void *addr)
{
	int ret = shmdt(addr);
	return ret < 0? -ERRNO_TO_PARA_ERROR(errno) : 1;
}
