/*
 * Copyright (C) 2006-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file write.h Writer-related structures. */

/** The list of supported writers. */
enum writer_enum {WRITER_ENUM};

/**
 * Describes one running instance of a writer.
 */
struct writer_node {
	/** The number of this writer. */
	int writer_num;
	/** Writer-specific data. */
	void *private_data;
	/** Pointer to the group this node belongs to. */
	struct writer_node_group *wng;
	/** The writer-specific configuration of this node. */
	void *conf;
	/** How much of the wng's buffer is already written. */
	size_t written;
};

/** Describes one supported writer. */
struct writer {
	/**
	 * The init function of the writer.
	 *
	 * It must fill in all other function pointers of the given
	 * writer structure.
	 */
	void (*init)(struct writer *w);
	/**
	 * The command line parser of the writer.
	 *
	 * It should check whether the command line options given by \a options are
	 * valid.  On success, it should return a pointer to the writer-specific
	 * configuration data determined by \a options.  Note that this might be called
	 * more than once with different values of \a options.
	 */
	void *(*parse_config)(const char *options);
	/**
	 * Open one instance of this writer.
	 *
	 * This function should perform any work necessary to write the incoming
	 * stream. To this aim, it may allocate its private data structure and store
	 * a pointer to that structure via the given writer_node parameter.
	 */
	int (*open)(struct writer_node *);
	/**
	 * Prepare the fd sets for select.
	 *
	 * This is called from the writer node group task's pre_select(). It
	 * may use the sched pointer to add any file descriptors or to decrease
	 * the select timeout. It must return positive on success and negative
	 * on errors.
	 */
	int (*pre_select)(struct sched *s, struct writer_node *wn);
	/**
	 * Write audio data.
	 *
	 * Called from the post_select function of the wng task. It must keep
	 * track of the number of bytes consumed from the wng's buffer via
	 * the \p wn->written variable (which may be modified by the wng handling
	 * functions). This function must return positive on success and
	 * negative on errors.
	 */
	int (*post_select)(struct sched *s, struct writer_node *wn);
	/**
	 * Close one instance of the writer.
	 *
	 * This function is assumed to succeed.
	 */
	void (*close)(struct writer_node *);
	/**
	 * Shutdown the writer
	 *
	 * This is a optional function pointer used for cleaning up.
	 */
	void (*shutdown)(struct writer_node *);
	struct ggo_help help;
};

/**
 * Describes a set of writer nodes that all write the same stream.
 */
struct writer_node_group {
	/** Number of nodes belonging to this group. */
	unsigned num_writers;
	/** Array of pointers to the corresponding writer nodes. */
	struct writer_node *writer_nodes;
	/** Non-zero if an error or end of file was encountered by the feeding task. */
	int *input_error;
	/** Current output buffer. */
	char **bufp;
	/** Number of bytes loaded in the output buffer. */
	size_t *loaded;
	/** Number of audio channels of the current stream. */
	unsigned int *channels;
	/** Sample rate of the current stream. */
	unsigned int *samplerate;
	/** The task associated to this group. */
	struct task task;
	/** Whether the group is open, i.e. wng_open() was called. */
	int open;
};

/** Loop over each writer node in a writer group. */
#define FOR_EACH_WRITER_NODE(i, wng) for (i = 0; i < (wng)->num_writers; i++)
/** Loop over each supported writer. */
#define FOR_EACH_WRITER(i) for (i = 0; i < NUM_SUPPORTED_WRITERS; i++)

/** Declare the init functions of all supported writers. */
DECLARE_WRITER_INITS;

/** Array containing the name of each writer. */
extern const char *writer_names[];

/** The writer structure for each supported writer. */
extern struct writer writers[NUM_SUPPORTED_WRITERS];
